/* pigeonhole-config.h.  Generated from pigeonhole-config.h.in by configure.  */

/* Define to the full name of Pigeonhole for Dovecot. */
#define PIGEONHOLE_NAME "Pigeonhole"

/* Define to the version of Pigeonhole for Dovecot. */
#define PIGEONHOLE_VERSION "0.5.21.1"

/* Pigeonhole ABI version */
#define PIGEONHOLE_ABI_VERSION "0.5.ABIv21(0.5.21.1)"

/* Define to build unfinished features/extensions. */
/* #undef HAVE_SIEVE_UNFINISHED */

/* LDAP support is built in */
/* #undef SIEVE_BUILTIN_LDAP */
